/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "data.h"

namespace bit {

  Data::Data( const void* d, size_t s, DataMode mode ) throw ( std::bad_alloc )
  {
    void * ptr;

    m_storage = Storage::pointer( new Storage() );

    switch ( mode ) {
      case COPY:
        m_storage->manage_data = true;
        ptr = ::malloc( s );
        if ( !ptr )
          throw std::bad_alloc();
        m_storage->size = s;
        m_storage->data = ( uint8_t* ) ptr;
        ::memcpy( m_storage->data, d, s );
        break;
      case MANAGED:
        m_storage->manage_data = true;
        m_storage->size = s;
        m_storage->data = ( uint8_t* ) d;
        break;
      case UNMANAGED:
        m_storage->manage_data = false;
        m_storage->size = s;
        m_storage->data = ( uint8_t* ) d;
        break;
    }

  }

  Data::Data( size_t s ) throw ( std::bad_alloc )
  {
    m_storage = Storage::pointer( new Storage() );

    m_storage->manage_data = true;

    if ( s != 0 )
    {
      void * ptr = ::calloc( s, 1 );
      if ( !ptr )
        throw std::bad_alloc();
      m_storage->size = s;
      m_storage->data = ( uint8_t* ) ptr;
    }
  }

  Data::Data( const Data& other ): m_storage( other.m_storage )
  {
  }

  Data::~Data()
  {
  }

  uint8_t* Data::data()
  {
    if ( ! m_storage ) return NULL;
    return m_storage->data;
  }

  const uint8_t* Data::data() const
  {
    if ( ! m_storage ) return NULL;
    return m_storage->data;
  }

  bool Data::set_data( const void* d, size_t s, DataMode mode ) throw ( std::bad_alloc )
  {
    uint8_t* ptr;

    // We don't deal with anything that overlaps the allocation
    if ( m_storage->data >= d && m_storage->data < (uint8_t*)d+s )
      return false;

    // if either the new data pointer is null or the size is zero, we have a request to clear the buffer
    if ( d == NULL || s == 0 ) {
      this->clear();
      return true;
    }

    switch ( mode ) {
      case COPY:

        // if we had external data, set the data pointer and size to null and 0
        // so that it will appear as if nothing has yet been allocated
        if ( ! m_storage->manage_data ) {
          m_storage->data = NULL;
          m_storage->size = 0;
        }

        // if it wasn't managed before, it's managed now
        m_storage->manage_data = true;

        // if we've reached this point we have a valid chunk of data to copy from.

        // We can always do a realloc, since a realloc with a null data buffer is equivalent to malloc.
        ptr = (uint8_t*) ::realloc( m_storage->data, s );
        if ( ptr == NULL )
          return false;
        m_storage->data = ptr;
        m_storage->size = s;
        ::memcpy( m_storage->data, d, s );
        break;
      case MANAGED:
        // if we had an allocation before, we need to free the memory
        if ( m_storage->data && m_storage->manage_data && m_storage->data != d )
        ::free( m_storage->data );

        m_storage->manage_data = true;
        m_storage->size = s;
        m_storage->data = ( uint8_t* ) d;
        break;
      case UNMANAGED:
        // if we had an allocation before, we need to free the memory
        if ( m_storage->data && m_storage->manage_data )
        ::free( m_storage->data );
        m_storage->manage_data = false;
        m_storage->size = s;
        m_storage->data = ( uint8_t* ) d;
        break;
    }

    return true;
  }

  size_t Data::size() const
  {
    return m_storage->size;
  }

  bool Data::resize( size_t s ) throw ( std::bad_alloc )
  {
    uint8_t* ptr;

    // nothing to do if the size is already the right size, so
    // we're successful
    if ( m_storage->size == s )
      return true;

    // if we have external data, we have to fail
    if ( ! m_storage->manage_data )
      return false;

    // if the request is to resize to 0, that's really a request to clear
    if ( s == 0 ) {
      this->clear();
      return true;
    }

    // At this point we really have to do some work
    // Now, realloc memory or if m_storage->data is NULL realloc will work like a malloc for us
    ptr = (uint8_t*) ::realloc( m_storage->data, s );

    // if the realloc failed we keep the memory the same and return false
    if ( ptr == NULL )
      return false;

    m_storage->data = ptr;

    // zero out newly allocated memory
    if ( s > m_storage->size )
      ::memset( m_storage->data + m_storage->size, 0x00, s - m_storage->size );

    m_storage->size = s;

    return true;
  }

  Data Data::clone() const
  {
    Data d( m_storage->data, m_storage->size, COPY );
    return d;
  }

  Data::operator bool()
  {
    return m_storage->data != NULL;
  }

  Data::operator bool() const
  {
    return m_storage->data != NULL;
  }

  Data::operator uint8_t*()
  {
    return m_storage->data;
  }

  Data::operator const uint8_t*() const
  {
    return m_storage->data;
  }

  std::string Data::hex_string( std::string separator ) const
  {
    std::string s;
    size_t i;
    char hex[ 3 ];

    hex[ 2 ] = 0x00;

    for ( i = 0; i < m_storage->size; i++ ) {
      if ( i != 0 )
        s += separator;
      hex[ 0 ] = '0' + ( m_storage->data[i] >> 4 );
      hex[ 1 ] = '0' + ( m_storage->data[i] & 0x0F );
      if ( hex[ 0 ] > '9' )
        hex[ 0 ] += 7;
      if ( hex[ 1 ] > '9' )
        hex[ 1 ] += 7;
      s += hex;
    }
    return s;
  }

  void Data::clear()
  {
      ::free( m_storage->data );
      m_storage->data = NULL;
      m_storage->size = 0;
      m_storage->manage_data = true;
  };

  bool Data::operator <( const Data& other ) const
  {
    return this->compare(other) < 0;
  }

  bool Data::operator <=( const Data& other ) const
  {
    return this->compare(other) <= 0;
  }

  bool Data::operator ==( const Data& other ) const
  {
    return this->compare(other) == 0;
  }

  bool Data::operator !=( const Data& other ) const
  {
    return this->compare(other) != 0;
  }

  bool Data::operator >=( const Data& other ) const
  {
    return this->compare(other) >= 0;
  }

  bool Data::operator >( const Data& other ) const
  {
    return this->compare(other) > 0;
  }

  int Data::compare( const Data& other ) const
  {
    if ( m_storage->size == 0 )
    {
      if ( other.m_storage->size == 0 ) return 0;
      return -1;
    }

    if ( other.m_storage->size == 0 ) return 1;

    // but if we made it here, both have storage
    size_t comp_size = std::min(m_storage->size, other.m_storage->size );

    int result = memcmp( m_storage->data, other.m_storage->data, comp_size );

    if ( result != 0 )
      return result;
    if ( m_storage->size < other.m_storage->size )
      return -1;
    if ( m_storage->size > other.m_storage->size )
      return 1;
    return 0;
  }

}
