/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "recordbuffer.h"

#include "recordvector.h"

namespace bit
  {

  RecordBuffer::RecordBuffer(size_t initial_size, size_t sizemax):
      Buffer(initial_size, sizemax),
      m_signal_field_changed_emitting(false)
  {}

  RecordBuffer::RecordBuffer(RecordBase::pointer record, size_t initial_size, size_t sizemax):
      Buffer(initial_size, sizemax),
      m_record(record),
      m_signal_field_changed_emitting(false)
  {}

  RecordBuffer::RecordBuffer(void* external_data, size_t data_octets, size_t sizemax, DataMode mode):
      Buffer(external_data, data_octets, sizemax, mode),
      m_signal_field_changed_emitting(false)
  {}

  RecordBuffer::RecordBuffer(RecordBase::pointer record, void* external_data, size_t data_octets, size_t sizemax, DataMode mode):
      Buffer(external_data, data_octets, sizemax, mode),
      m_record(record),
      m_signal_field_changed_emitting(false)
  {}

  RecordBuffer::pointer RecordBuffer::create(size_t initial_size, size_t sizemax) {
    return pointer( new RecordBuffer(initial_size, sizemax) );
  }

  RecordBuffer::pointer RecordBuffer::create(RecordBase::pointer record, size_t initial_size, size_t sizemax) {
    return pointer( new RecordBuffer( record, initial_size, sizemax ) );
  }

  RecordBuffer::pointer RecordBuffer::create(void* external_data, size_t data_octets, size_t sizemax, DataMode mode ) {
    return pointer( new RecordBuffer( external_data, data_octets, sizemax, mode ) );
  }

  RecordBuffer::pointer RecordBuffer::create(RecordBase::pointer record,
                        void* external_data,
                        size_t data_octets,
                        size_t sizemax,
                        DataMode mode
                       )
  {
    return pointer( new RecordBuffer( record, external_data, data_octets, sizemax, mode ) );
  }

  RecordBuffer::~RecordBuffer()
  {}

  const std::string & RecordBuffer::xml( )
  {
    m_xml = this->xml( m_record );
    return m_xml;
  }

  std::string RecordBuffer::xml(FieldBase::pointer f) {
    std::ostringstream sout;

    if ( f->container_type() == FIELD || f->container_type() == FIELD_ELEMENT )
      return FieldBuffer(*this, f).xml();

    if ( f->container_type() == RECORD || f->container_type() == RECORD_VECTOR || f->container_type() == RECORD_ELEMENT ) {
      sout << "<data_record name=\"" << f->name();
      if ( f->container_type() == RECORD_ELEMENT )
        sout << "[" << bit_dynamic_pointer_cast<RecordVector::Element>(f)->index() << "]";
      sout << "\">\n";
    }

    for (size_t i = 0; i < f->fields(); i++)
      sout << this->xml( f->field(i) ) << "\n";

    if ( f->container_type() == RECORD || f->container_type() == RECORD_VECTOR )
      sout << "</data_record>\n";

    return sout.str();
  }

  FieldBuffer RecordBuffer::field(size_t field_id) throw (exception::no_record)
  {
    if (m_record)
      return FieldBuffer(*this, m_record->field(field_id));
    else
      throw exception::no_record();
  }

  FieldBuffer RecordBuffer::field(std::string field_id) throw (exception::no_record)
  {
    if (m_record)
      return FieldBuffer(*this, m_record->field(field_id));
    else
      throw exception::no_record();
  }

  FieldBuffer RecordBuffer::field(FieldBase::iterator iter)
  {
    return FieldBuffer( *this, iter.pointer() );
  }

  FieldBuffer RecordBuffer::field(FieldBase::pointer pointer)
  {
    return FieldBuffer( *this, pointer );
  }

  FieldBuffer RecordBuffer::operator[](size_t field_id) throw (exception::no_record)
  {
    return this->field(field_id);
  }

  FieldBuffer RecordBuffer::operator[](std::string field_id) throw (exception::no_record)
  {
    return this->field(field_id);
  }

  FieldBuffer RecordBuffer::operator[](const char* field_id) throw (exception::no_record)
  {
    return this->field(field_id);
  }

  FieldBuffer RecordBuffer::operator[](FieldBase::iterator iter)
  {
    return this->field(iter);
  }

  FieldBuffer RecordBuffer::operator[](FieldBase::pointer pointer)
  {
    return this->field(pointer);
  }

  void RecordBuffer::set_record(RecordBase::pointer record ) throw ()
  {
    m_record = record;
    this->on_record_changed();
  }

  void RecordBuffer::clear_record() throw ()
  {
    m_record.reset();
    this->on_record_changed();
  }

  RecordBase::pointer RecordBuffer::record( )
  {
    return m_record;
  }

  sigc::signal< void, FieldBuffer > RecordBuffer::signal_field_changed( )
  {
    return m_signal_field_changed;
  }

  void RecordBuffer::on_field_changed( FieldBase::pointer field )
  {
    FieldBase::pointer current = field;

    if (m_signal_field_changed_emitting)
      {
        //we only need to mark each field for emitting once, so check to see if we already
        //have it queued for emitting
        if (m_signal_field_changed_set.find(current) == m_signal_field_changed_set.end())
          {
            //it's not already marked, so add it to both data structures
            m_signal_field_changed_set.insert( current );
            m_signal_field_changed_queue.push( current );
          }
        return;
      }

    m_signal_field_changed_emitting = true;
    do
      {
        if (m_signal_field_changed_set.size() > 0)
          {
            current = m_signal_field_changed_queue.front();
            m_signal_field_changed_queue.pop();
            m_signal_field_changed_set.erase(current);
          }
          m_signal_field_changed.emit( FieldBuffer(*this, current) );
      }
    while (m_signal_field_changed_set.size() > 0);

    m_signal_field_changed_emitting = false;
    return;
  }

  void RecordBuffer::on_record_changed( )
  {
    m_signal_record_changed.emit();
  }

  sigc::signal<void> RecordBuffer::signal_record_changed( )
  {
    return m_signal_record_changed;
  }

  void RecordBuffer::set_data( const void* data, size_t size )
  {
    Buffer::set_data(data, size);
    m_signal_record_changed.emit();
  }


}
